/*
 * Routine conu par Raphal Pasquier
*/
/*
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Multiplication de 2 nombres par la mthode de Pollard
 * Version optimise pour la vitesse
 */

/* dfini dans lf32.c */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <math.h>
//#include "op_fftp.h"

#define BASE32    1000000000
#define BASE32_BIN 1073741824
#define C1        (double) 0.22156187299249082 // ln(10) / ln(2) / 15
#define C1P       (double) 0.23728057820624017 // ln(10) / ln(2) / 15
#define CACHE32     999999999L
#define CACHE32_BIN     1073741823L

#define NONBORNE  3

#define DECALAGE32_BIN  30

#define INFTY_DEPTH -1000000.0

#define CACHE_BASE16  15L
#define PARITE        1L
#define PI 3.14159265358979323846

/*  si la norme N1 de z(n) est suprieure  3, alors on est sr que
    la suite complexe dfinie par z(n+1) = z(n)^2 + z(0) est
    divergente en norme euclidienne.      */

typedef struct
    {
      int  signe ;
      long taille_mantisse ;
      long* mantisse ;
} Li ;

typedef struct
    {
     int signe;
     long indiceMax ;
     long taille_mantisse ; // nombre de dcimales aprs la virgule.
     long* mantisse ;
} Lii ;


void createLi( Li* li , long nb ) ;
void destroyLi( Li* li ) ;
long indiceMax( long* chiffre , Li* a ) ;
void printLi( Li* li ) ;
void copy( Li* nouveau , Li* li ) ;
void copyAndCreate( Li* nouveau , Li* li ) ;
void permutLi( Li* li2 , Li* li ) ; // je suppose que li et li2 ont mme taille

//-----------------------------------------------------------------------------
// ----------------   Entier long en base 10^9        -------------------------
//-----------------------------------------------------------------------------

long div_R_deci( Li* li , long D ) ;
long divB_deci( Li* res , Li* li , long D ) ;
//Li create_BigInteger_deci( char* BigInteger ) ;
int create_BigFloat_deci( Li* bf , char* BigFloat ) ;
int create_BigFloat_deci2( Li* bf , long taille , char* BigFloat ) ;
void mulB_deci( Li* prod , Li* li , long m ) ;
//long div_by_pow2_deci( Li* result , Li* li , long p ) ;
long addB2_R_deci( Li* a , long m ) ;
long mulB2_R_deci( Li* al , long m  ) ;
double depth_setMandelbrot(double xInf, double xSup, double yInf, double ySup ) ;
double depth_deci( Li* bf ) ;
double depth_setMandelbrot_deci(char* xInf , char* xSup , char* yInf ,char* ySup ) ;
void add_2P_deci( Li* result , Li* a , Li* b ) ;
void sub_2P_deci( Li* result , Li* a , Li* b ) ;
void add_R_2P_deci( Li* a , Li* b ) ;
void sub_R_2P_deci( Li* a , Li* b ) ;
void mul_fft_Pollard_deci_trunc( Li* result , Li* a , Li* b , long k , unsigned long* prod ,
		    unsigned long* tp0, unsigned long* tp1, unsigned long* tp2 ) ;
void square_2P_trunc_deci( Li* prod , Li* a ) ;
void square_deci( Li* result , Li* prod , Li* a ) ;
void add_deci( Li* result , Li* a , Li* b ) ;
void sub_deci( Li* result , Li* a , Li* b ) ;
void add_R_deci( Li* a , Li* b ) ;
void sub_R_deci( Li* a , Li* b ) ;
char* bigFloat_deciToStr( Li* li ) ;

//-----------------------------------------------------------------------------
// ----------------   Entier long en base 2^30      ---------------------------
//-----------------------------------------------------------------------------

void add_R_2P_bin( Li* a , Li* b ) ;
void add_2P_bin( Li* result , Li* a , Li* b ) ;
void sub_2P_bin( Li* result , Li* a , Li* b ) ;
void sub_R_2P_bin( Li* a , Li* b ) ;
void add_bin( Li* result , Li* a , Li* b ) ;
void add_R_bin( Li* a , Li* b ) ;
void sub_bin( Li* result , Li* a , Li* b ) ;
void sub_R_bin( Li* a , Li* b ) ;
void mulB_bin( Li* prod , Li* li , long m ) ;// suppose que prod est dj cr
long div_by_pow2_bin( Li* result , Li* li , long p ) ;// suppose que result est dj cr
void mul_2P_bin( Li* prod , Li* a , Li* b ) ;// suppose que prod est dj cr
void mul_bin( Li* result , Li* prod , Li* a , Li* b ) ;// suppose que prod, result sont dj crs
void square_2P_bin( Li* prod , Li* a ) ;
void square_2P_trunc_bin( Li* prod , Li* a ) ;
void square_bin( Li* result , Li* prod , Li* a ) ;
void mul_by2_R_bin( Li* a ) ;
double BigFloatToFloat( Li* bfBaseBin ) ;
void BigFloat_deciToBigFloat( Li* bf , Li* bfBaseBin ) ;// suppose que bfBaseBin est dj cr


//-----------------------------------------------------------------------------
void createLii( Lii* li , long nb ) ;
void destroyLii( Lii* li ) ;
void copyLii(  Lii* nouveau ,Lii* li ) ;
void printLii( Lii* li ) ;
//-----------------------------------------------------------------------------
// -------------   Entier long en base 10^9  avec indiceMax     ---------------
//-----------------------------------------------------------------------------

int initialise_fixedPoint_decii( Lii* le , long e ) ;
int initialise_int_decii( Lii* le , long e ) ;
long add_R_2P_decii_spe( Lii* a , Lii* b ) ;
void sub_R_2P_decii_spe( Lii* a , Lii* b ) ;
long divB_R_decii( Lii* a , long D ) ;
long divB_decii( Lii* res , Lii* a , long D ) ;
long mulB_R_decii( Lii* a , long m ) ;
long mul_2P_decii( Lii* prod , Lii* a , Lii* b ) ;
void print_cst_base10_decii( Lii* a ) ;

//-----------------------------------------------------------------------------
// -------------   Entier long en base 2^30  avec indiceMax     ---------------
//-----------------------------------------------------------------------------

int initialise_fixedPoint_bini( Lii* le , long e ) ;
int initialise_int_bini( Lii* le , long e ) ;
long add_R_2P_bini_spe( Lii* a , Lii* b ) ;
void sub_R_2P_bini_spe( Lii* a , Lii* b ) ;
long divB_R_bini( Lii* a , long D ) ;
long divB_bini( Lii* res , Lii* a , long D ) ;
long mulB_R_bini( Lii* a , long m ) ;
long mul_2P_bini( Lii* prod , Lii* a , Lii* b ) ;
void print_cst_base2_bini( Lii* a ) ;
void print_cst_base16_bini( Lii* a ) ;

//-----------------------------------------------------------------------------
// -------------   Entier long en base 2^30  de bas niveau     ----------------
//-----------------------------------------------------------------------------

void mul_2P_naif_bin( long* prod , long* a , long* b , long n ) ;
void square_2P_naif_bin( long* prod , long* a , long n ) ;





long** createMatrix( long n ) ;
void destroyMatrix( long** ma , long n ) ;

long calcul_germe_mandelbrotNorm1(Li* BinRe , Li* BinIm , Li* BinCx , Li* BinCy , long iter,
			     Li* a0 , Li* a1 , Li* a2 , Li* prod , long tailleBin ) ;
long calcul_germe_mandelbrot(Li* BinRe , Li* BinIm , Li* BinCx , Li* BinCy , long iter,
			     Li* a0 , Li* a1 , Li* a2 , Li* prod , long tailleBin ) ;

int calcul_MandelbrotNorm1( char* xInf, char* xSup, char* yInf, char* ySup, long tailleBin,long iter ,
                            long **hypert ) ;
int calcul_Mandelbrot( char* xInf, char* xSup, char* yInf, char* ySup, long tailleBin,long iter ,
                       long **hypert ) ;
int calcul_Mandelbrot2( char* xInf ,char* xSup ,char* yInf ,char* ySup ,
		           long tailleBin,long iter ,long **hypert, long sizeMatrix ) ;
int calcul_Mandelbrot_1Quart( char* xInf, char* xSup, char* yInf, char* ySup, long tailleBin,long iter ,
                              long **hypert , long corner ) ;

		       /*
int utilisation_Mandelbrot( char* xInf , char* xSup , char* yInf ,
                              char* ySup , long iter , char* name  ) ;
int utilisation_Mandelbrot_1Quart( char* xInf , char* xSup , char* yInf ,
                              char* ySup , long iter , char* name  ) ;

int utilisation_Mandelbrot_9_v2( char* xInf , char* xSup , char* yInf ,
                              char* ySup , long iter , char* name  ) ;
*/

