/*
 * Routine conu par Fabrice Bellard
*/
/*
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * Multiplication de 2 nombres par la mthode de Pollard
 * Version optimise pour la vitesse
 */

//#include <stdlib.h>
//#include <stdio.h>
#include <math.h>
//#include "op_fftp.h"
#include "lf32.h"


/*
 * calcul de K=(y0+m0*(y1+m1*y2))+a0+a1*2^32
 * a1:a0=K/base
 * return K % base
 */

unsigned long BaseCvt(unsigned long *a0,unsigned long *a1,
		      unsigned long y0,unsigned long y1,unsigned long y2,unsigned long m0,unsigned long m1)
{
	 unsigned long long  t1,t2,t3,t4;

	 t1=y1+(unsigned long long )m1*(unsigned long long )y2;
	 t2=(t1 & 0xFFFFFFFFL) * (unsigned long long )m0 + y0 + *a0;
	 t3=(t1>>32) * (unsigned long long )m0 + (t2>>32) + *a1;

	 *a1=t3/BASE32 ;
	 t4=((t3 % BASE32 )<<32)+(t2 & 0xFFFFFFFFL);
	 *a0=t4 / BASE32 ;

	 return t4 % BASE32 ;
}

unsigned long BaseCvt_bin(unsigned long *a0,unsigned long *a1,
		          unsigned long y0,unsigned long y1,unsigned long y2,unsigned long m0,unsigned long m1)
{
	 unsigned long long  t1,t2,t3,t4;

	 t1=y1+(unsigned long long )m1*(unsigned long long )y2;
	 t2=(t1 & 0xFFFFFFFFL) * (unsigned long long )m0 + y0 + *a0;
	 t3=(t1>>32) * (unsigned long long )m0 + (t2>>32) + *a1;

	 *a1=t3 >> DECALAGE32_BIN ;
	 t4=((t3 & CACHE32_BIN )<<32)+(t2 & 0xFFFFFFFFL);
	 *a0=t4 >> DECALAGE32_BIN ;

	 return t4 & CACHE32_BIN ;
}
/*
 * Oprations modulaires
 */

/*
inline unsigned long add_mod(unsigned long a,unsigned long b,unsigned long modulo)  {
	 unsigned long c;
	 c=a+b;
	 if (c>=modulo) c-=modulo;
	 return c;
}

inline unsigned long sub_mod(unsigned long a,unsigned long b,unsigned long modulo)  {
	 unsigned long c;
	 if (a>=b) {
			c=a-b;
	 } else {
			c=a-b+modulo;
	 }
	 return c;
}
*/

inline unsigned long add_mod(unsigned long a,unsigned long b,unsigned long modulo)  {
	 int c,d;
	 c=a+b-modulo;
	 d=c >> 31;
	 c=c + (d & modulo);
	 return c;
}

inline unsigned long sub_mod(unsigned long a,unsigned long b,unsigned long modulo)  {
	 int c,d;
	 c=a-b;
	 d=c >> 31;
	 c=c + (d & modulo);
	 return c;
}


#ifndef ASM_386

inline unsigned long mul_mod(unsigned long a,unsigned long b,unsigned long modulo)  {
	 unsigned long c;
	 c=((unsigned long long )a*(unsigned long long )b) % modulo;
	 return c;
}

#else

inline unsigned long mul_mod(unsigned long a,unsigned long b,unsigned long modulo)  {
	 unsigned long c;
	 asm(" movl %1,%%eax ; mull %2 ; divl %3 ; movl %%edx,%0 "
			 : "=g" (c)
			 : "g" (a) , "g" (b) , "g" (modulo)
			 : "%eax" , "%edx"
			 );
	 return c;
}

#endif

/*
 * inversion de x mod y
 */
unsigned long inv_mod(unsigned long x,unsigned long y) {
	 unsigned long q,u,v,a,c,t;
	 u=x;
	 v=y;
	 c=1;
	 a=0;
	 while (u!=0) {
			q=v/u;

			t=c;
			c=sub_mod(a,mul_mod(q,c,y),y);
			a=t;
			
			t=u;
			u=v-q*u;
			v=t;
	 }
	 return a;
}

#define PRIM_PRECALC_BITS  9
#define PRIM_PRECALC_NB     (1<<PRIM_PRECALC_BITS)


/* FFT Pass m on 2^n unsigned long */
static void FFTPass_NoPre(int n,int m,unsigned long *tab,unsigned long *tabpow,unsigned long modulo)
{
	 int i,j,k,l;
	 unsigned long a,b,*p,*q,cpow,mpow;
	 
	 mpow=tabpow[n-m];
	 k=1 << (n-m-1);
	 l=1 << m;

	 p=tab;
	 q=tab+k;
	 for(i=0;i<l;i++)  {
				 
			a=*p;
			b=*q;
			*p=add_mod(a,b,modulo);
			*q=sub_mod(a,b,modulo);
			p++;
			q++;
				 
			cpow=mpow;
			j=k-1;
			while(1) {
				 a=*p;
				 b=*q;
				 *p=add_mod(a,b,modulo);
				 *q=mul_mod(sub_mod(a,b,modulo),cpow,modulo);
				 p++;
				 q++;
				 j--;
				 if (j==0) break;
				 cpow=mul_mod(cpow,mpow,modulo);
			}

			p+=k;
			q+=k;
	 }
}

/* idem but with cpow precalculated */

static void FFTPass_Pre(int n,int m,unsigned long *tab,unsigned long *tabpowpre,unsigned long modulo) 
{
	 int i,j,k,l,s;
	 unsigned long a,b,*p,*q,*r;
	 
	 k=1 << (n-m-1);
	 l=1 << m;
	 s=1 << (PRIM_PRECALC_BITS+m+1-n);
	 
	 p=tab;
	 q=tab+k;
	 for(i=0;i<l;i++)  {
				 
				 a=*p;
				 b=*q;
				 *p=add_mod(a,b,modulo);
				 *q=sub_mod(a,b,modulo);
				 p++;
				 q++;

				 r=tabpowpre+s;
				 j=k-1;
				 do {
						a=*p;
						b=*q;
						*p=add_mod(a,b,modulo);
						*q=mul_mod(sub_mod(a,b,modulo),*r,modulo);
						p++;
						q++;
						r+=s;
				 } while (--j!=0);

				 p+=k;
				 q+=k;
	 }
}

/* last pass */

static void FFTPass_Last(int n,unsigned long *tab,unsigned long modulo)
{
	 int i,l;
	 unsigned long a,b,*p;

	 p=tab;
	 l=1<<(n-1);
	 for(i=0;i<l;i++) {
			a=p[0];
			b=p[1];
			p[0]=add_mod(a,b,modulo);
			p[1]=sub_mod(a,b,modulo);
			p+=2;
	 }
}


/* 
 * FFT sur 2**n unsigned long 
 * Note: Le rsultat est dans un tableau  parcourir avec un index  bits
 * inverss
 */
void FFT(int n,unsigned long *tab,unsigned long *tabpowpre,unsigned long *tabpow,unsigned long modulo)
{
	 int pass,pass1,pass2;
	 
	 pass1=n-1-PRIM_PRECALC_BITS;
	 if (pass1<0) pass1=0;
	 pass2=n-1;

	 for(pass=0;pass<pass1;pass++) 
	 	FFTPass_NoPre(n,pass,tab,tabpow,modulo);

	 for(pass=pass1;pass<pass2;pass++) 
	 	FFTPass_Pre(n,pass,tab,tabpowpre,modulo);
	 
	 FFTPass_Last(n,tab,modulo);
}

/*
 * Prcalcul du tableau d'inversion de taille 2**k
 */

unsigned long *tabrev=NULL;
unsigned long tabrev_k=0;

void PreReverse(unsigned long k) {
	 unsigned long a,c,i,j,nb_word;

	 if (k!=tabrev_k) {
			nb_word=1 << k;
			tabrev= (unsigned long*) realloc( tabrev, nb_word * sizeof(unsigned long) );
			tabrev_k=k;
			for(i=0;i<nb_word;i++) {
				 a=i;
				 c=0;
				 for(j=0;j<k;j++) {
						c=(c<<1) | (a & 1);
						a=a >> 1;
				 }
				 tabrev[i]=c;
			}
	 }
}


/*
 * Reconstruction d'un tableau de taille 2**k en renversant l'ordre des
 * bits de l'index.
 */

void FFTReverse(unsigned long k,unsigned long *tab) {
	 unsigned long i,j,m,mask,mask1,nb_word;
	 unsigned long tmp;

	 nb_word=1 << k;
	 m=k / 2;
	 PreReverse(m);
	 mask= (1 << m) - 1;

	 if ( (k&1)==0 ) {
			/* cas o n est pair */
			for(i=1;i<nb_word;i++)  {
				 j=(tabrev[i & mask] << m) + tabrev[i>>m];
				 if (j<i) {
						tmp=tab[i];
						tab[i]=tab[j];
						tab[j]=tmp;
				 }
			}
	 } else {
			/* cas o n est impair */
			mask1=1 << m;
			m++;
			for(i=1;i<nb_word;i++)  {
				 j=(tabrev[i & mask] << m) + tabrev[i>>m] + (i & mask1);
				 if (j<i) {
						tmp=tab[i];
						tab[i]=tab[j];
						tab[j]=tmp;
				 }
			}
	 }
}

/*
 * Etape a effectuer avant la FFT inverse: tab[u]<->tab[N-u] (u=1..N/2)
 */
void FFTNegate(unsigned long k,unsigned long *tab) {
	 unsigned long i,n,m,tmp;

	 m=1 << (k-1);
	 n=1 << k;
	 for(i=1;i<m;i++) {
			tmp=tab[i];
			tab[i]=tab[n-i];
			tab[n-i]=tmp;
	 }
}

/*
 * Implmentation de la multiplication
 */

#define NB_FFTWORD 3
#define PRIM_ORDER 25

/* nombres premiers pour le reste chinois, dans l'ordre croissant */
unsigned long prem[NB_FFTWORD]= {
	 1711276033, /* 51*2^25+1 */
	 1811939329, /* 27*2^26+1 */
	 2013265921, /* 15*2^27+1 */
};

/* racines principales de l'unit d'ordre PRIM_ORDER */
unsigned long prim[NB_FFTWORD]= {
	 40,
	 103,
	 170
};

/* puissances du type 2^k des racines principales */
unsigned long tabpow[NB_FFTWORD][PRIM_ORDER+1];
/* prcalcul des puissances des racines principales */
unsigned long tabpowpre[NB_FFTWORD][PRIM_PRECALC_NB];

/*
 * Init FFT multiplicatin by Pollard
 */

void init_mul_fft_Pollard(void) {
	 unsigned long a,b,*p;
	 int i,j;

	 /* Initialisation des puissances des racines principales */

	 for(i=0;i<NB_FFTWORD;i++) {
			a=prim[i];
			for(j=PRIM_ORDER;j>=1;j--) {
				 tabpow[i][j]=a;
				 a=mul_mod(a,a,prem[i]);
			}
			a=tabpow[i][PRIM_PRECALC_BITS+1];
			b=1;
			p=&tabpowpre[i][0];
			for(j=0;j<PRIM_PRECALC_NB;j++) {
				 *p++=b;
				 b=mul_mod(b,a,prem[i]);
			}
	 }
}

/*
 * Multiplication de 2 nombres de 2^k mots. Rsultat dans un nombre de
 * 2^(k+1) mots.
 * Si op1==op2, alors un carr est calcul.
 */

void mul_fft_Pollard_deci(int k,unsigned long *result,unsigned long *op1,unsigned long *op2,
	    unsigned long* tmp0, unsigned long* tmp1, unsigned long* tmp2 )

{
 long res_size,op_size;
 unsigned long i,j;
 unsigned long y[NB_FFTWORD],u[NB_FFTWORD],c01,c02,c12;
 unsigned long a0,a1,a;

 res_size = 1 << (k+1) ;
 op_size = 1 << k ;

 /* Calcul des NB_FFTWORD convolutions */
 // tmp0
 /* stockage des 2 nombres */
 for(j = 0; j < op_size ; j++ ) tmp0[j] = op1[j] % prem[0] ;
 for(j = op_size ; j < res_size ; j++ ) tmp0[j] = 0 ;

 if (op2 != op1)
	{
	 for(j=0;j<op_size;j++) result[j] = op2[j] % prem[0] ;
	 for(j=op_size;j<res_size;j++) result[j] = 0 ;
	}
 /* calcul de la convolution de op1 & op2 par FFT */
 FFT(k+1,tmp0,tabpowpre[0],tabpow[0],prem[0]);

 if (op1 != op2)
	{
	 FFT(k+1,result,tabpowpre[0],tabpow[0],prem[0]) ;
	 for(j=0;j<res_size;j++) tmp0[j] = mul_mod(tmp0[j],result[j],prem[0]) ;
	}
	else
	{
	 for(j=0;j<res_size;j++)
		{
		 a = tmp0[j] ;
		 tmp0[j] = mul_mod(a,a,prem[0]) ;
		}
	}

 FFTReverse(k+1,tmp0) ;
 FFTNegate(k+1,tmp0) ;
 FFT(k+1,tmp0,tabpowpre[0],tabpow[0],prem[0]) ;
 FFTReverse(k+1,tmp0);

 // tmp1
 /* stockage des 2 nombres */
 for(j=0;j<op_size;j++) tmp1[j] = op1[j] % prem[1] ;
 for(j=op_size;j<res_size;j++)  tmp1[j] = 0 ;

 if (op2!=op1)
	{
	 for(j=0;j<op_size;j++) result[j] = op2[j] % prem[1] ;
	 for(j=op_size;j<res_size;j++) result[j] = 0 ;
	}
 /* calcul de la convolution de op1 & op2 par FFT */
 FFT(k+1,tmp1,tabpowpre[1],tabpow[1],prem[1]) ;

 if (op1!=op2)
	{
	 FFT(k+1,result,tabpowpre[1],tabpow[1],prem[1]);
	 for(j=0;j<res_size;j++) tmp1[j] =mul_mod(tmp1[j],result[j],prem[1]) ;
	}
	else
	{
	 for(j=0;j<res_size;j++)
		{
		 a = tmp1[j] ;
		 tmp1[j] = mul_mod(a,a,prem[1]) ;
		}
	}

 FFTReverse(k+1,tmp1) ;
 FFTNegate(k+1,tmp1) ;
 FFT(k+1,tmp1,tabpowpre[1],tabpow[1],prem[1]) ;
 FFTReverse(k+1,tmp1);

 // tmp2
 /* stockage des 2 nombres */
 for(j=0;j<op_size;j++) tmp2[j] = op1[j] % prem[2] ;
 for(j=op_size;j<res_size;j++) tmp2[j] = 0;

 if (op2!=op1)
	{
	 for(j=0;j<op_size;j++) result[j]= op2[j] % prem[2] ;
	 for(j=op_size;j<res_size;j++) result[j] = 0 ;
	}

 /* calcul de la convolution de op1 & op2 par FFT */

 FFT(k+1,tmp2,tabpowpre[2],tabpow[2],prem[2]);

 if (op1 != op2)
	{
	 FFT(k+1,result,tabpowpre[2],tabpow[2],prem[2]) ;
	 for(j=0;j<res_size;j++) tmp2[j] = mul_mod(tmp2[j],result[j],prem[2]);
	}
	else
	{
	 for(j=0;j<res_size;j++)
		{
		 a = tmp2[j] ;
		 tmp2[j] = mul_mod(a,a,prem[2]);
		}
	}

 FFTReverse(k+1,tmp2);
 FFTNegate(k+1,tmp2);
 FFT(k+1,tmp2,tabpowpre[2],tabpow[2],prem[2]);
 FFTReverse(k+1,tmp2);

 /* rassemblement des NB_FFTWORD convolutions, reste
  * chinois, et propagation des retenues
  * on a "hardcod" NB_FFTWORD=3
  */
 for(i=0;i<NB_FFTWORD;i++) u[i]=inv_mod(1<<(k+1),prem[i]);
 c01=inv_mod(prem[0],prem[1]);
 c02=inv_mod(prem[0],prem[2]);
 c12=inv_mod(prem[1],prem[2]);

 /* init retenues */
 a0=a1=0;

 for(j=0;j<res_size;j++)
	{
	 /* division par 2^(k+1) */

	 y[0]=mul_mod(u[0],tmp0[j],prem[0]);
	 y[1]=mul_mod(u[1],tmp1[j],prem[1]);
	 y[2]=mul_mod(u[2],tmp2[j],prem[2]);

	 /* calcul de la reprsentation "mixed radix" */

	 y[1]=mul_mod(sub_mod(y[1],y[0],prem[1]),c01,prem[1]);
	 y[2]=mul_mod(sub_mod(y[2],y[0],prem[2]),c02,prem[2]);
	 y[2]=mul_mod(sub_mod(y[2],y[1],prem[2]),c12,prem[2]);

	 /* conversion dans la base de sortie */

	 result[j]=BaseCvt(&a0,&a1,y[0],y[1],y[2],prem[0],prem[1]);

	 }
}

void mul_fft_Pollard_bin(int k,unsigned long *result,unsigned long *op1,unsigned long *op2,
	        unsigned long* tmp0, unsigned long* tmp1, unsigned long* tmp2 )

{
 long res_size,op_size;
 unsigned long i,j;
 unsigned long y[NB_FFTWORD],u[NB_FFTWORD],c01,c02,c12;
 unsigned long a0,a1,a;

 res_size = 1 << (k+1) ;
 op_size = 1 << k ;

 /* Calcul des NB_FFTWORD convolutions */
 // tmp0
 /* stockage des 2 nombres */
 for(j = 0; j < op_size ; j++ ) tmp0[j] = op1[j] % prem[0] ;
 for(j = op_size ; j < res_size ; j++ ) tmp0[j] = 0 ;

 if (op2 != op1)
	{
	 for(j=0;j<op_size;j++) result[j] = op2[j] % prem[0] ;
	 for(j=op_size;j<res_size;j++) result[j] = 0 ;
	}
 /* calcul de la convolution de op1 & op2 par FFT */
 FFT(k+1,tmp0,tabpowpre[0],tabpow[0],prem[0]);

 if (op1 != op2)
	{
	 FFT(k+1,result,tabpowpre[0],tabpow[0],prem[0]) ;
	 for(j=0;j<res_size;j++) tmp0[j] = mul_mod(tmp0[j],result[j],prem[0]) ;
	}
	else
	{
	 for(j=0;j<res_size;j++)
		{
		 a = tmp0[j] ;
		 tmp0[j] = mul_mod(a,a,prem[0]) ;
		}
	}

 FFTReverse(k+1,tmp0) ;
 FFTNegate(k+1,tmp0) ;
 FFT(k+1,tmp0,tabpowpre[0],tabpow[0],prem[0]) ;
 FFTReverse(k+1,tmp0);

 // tmp1
 /* stockage des 2 nombres */
 for(j=0;j<op_size;j++) tmp1[j] = op1[j] % prem[1] ;
 for(j=op_size;j<res_size;j++)  tmp1[j] = 0 ;

 if (op2!=op1)
	{
	 for(j=0;j<op_size;j++) result[j] = op2[j] % prem[1] ;
	 for(j=op_size;j<res_size;j++) result[j] = 0 ;
	}
 /* calcul de la convolution de op1 & op2 par FFT */
 FFT(k+1,tmp1,tabpowpre[1],tabpow[1],prem[1]) ;

 if (op1!=op2)
	{
	 FFT(k+1,result,tabpowpre[1],tabpow[1],prem[1]);
	 for(j=0;j<res_size;j++) tmp1[j] =mul_mod(tmp1[j],result[j],prem[1]) ;
	}
	else
	{
	 for(j=0;j<res_size;j++)
		{
		 a = tmp1[j] ;
		 tmp1[j] = mul_mod(a,a,prem[1]) ;
		}
	}

 FFTReverse(k+1,tmp1) ;
 FFTNegate(k+1,tmp1) ;
 FFT(k+1,tmp1,tabpowpre[1],tabpow[1],prem[1]) ;
 FFTReverse(k+1,tmp1);

 // tmp2
 /* stockage des 2 nombres */
 for(j=0;j<op_size;j++) tmp2[j] = op1[j] % prem[2] ;
 for(j=op_size;j<res_size;j++) tmp2[j] = 0;

 if (op2!=op1)
	{
	 for(j=0;j<op_size;j++) result[j]= op2[j] % prem[2] ;
	 for(j=op_size;j<res_size;j++) result[j] = 0 ;
	}

 /* calcul de la convolution de op1 & op2 par FFT */

 FFT(k+1,tmp2,tabpowpre[2],tabpow[2],prem[2]);

 if (op1 != op2)
	{
	 FFT(k+1,result,tabpowpre[2],tabpow[2],prem[2]) ;
	 for(j=0;j<res_size;j++) tmp2[j] = mul_mod(tmp2[j],result[j],prem[2]);
	}
	else
	{
	 for(j=0;j<res_size;j++)
		{
		 a = tmp2[j] ;
		 tmp2[j] = mul_mod(a,a,prem[2]);
		}
	}

 FFTReverse(k+1,tmp2);
 FFTNegate(k+1,tmp2);
 FFT(k+1,tmp2,tabpowpre[2],tabpow[2],prem[2]);
 FFTReverse(k+1,tmp2);

 /* rassemblement des NB_FFTWORD convolutions, reste
  * chinois, et propagation des retenues
  * on a "hardcod" NB_FFTWORD=3
  */
 for(i=0;i<NB_FFTWORD;i++) u[i]=inv_mod(1<<(k+1),prem[i]);
 c01=inv_mod(prem[0],prem[1]);
 c02=inv_mod(prem[0],prem[2]);
 c12=inv_mod(prem[1],prem[2]);

 /* init retenues */
 a0=a1=0;

 for(j=0;j<res_size;j++)
	{
	 /* division par 2^(k+1) */

	 y[0]=mul_mod(u[0],tmp0[j],prem[0]);
	 y[1]=mul_mod(u[1],tmp1[j],prem[1]);
	 y[2]=mul_mod(u[2],tmp2[j],prem[2]);

	 /* calcul de la reprsentation "mixed radix" */

	 y[1]=mul_mod(sub_mod(y[1],y[0],prem[1]),c01,prem[1]);
	 y[2]=mul_mod(sub_mod(y[2],y[0],prem[2]),c02,prem[2]);
	 y[2]=mul_mod(sub_mod(y[2],y[1],prem[2]),c12,prem[2]);

	 /* conversion dans la base de sortie */

	 result[j]=BaseCvt_bin(&a0,&a1,y[0],y[1],y[2],prem[0],prem[1]);

	 }
}




