/*
 * Routine conu par Raphal Pasquier
*/
/*
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "lf32.h"
#include "op_fftw3.h"
#include "fft_common.h"
#include <fftw3.h>

// Cette routine permet l'initialisation des varaibles ncessaires  la
// multiplication de deux entiers de type Li de taille 2^k en base 2^30.
void init_mul_fftw3( long k, fftw_plan* pA , fftw_plan* pB , fftw_plan* p_i ,
		     fftw_complex** inoutA , fftw_complex** inoutB )
{
 long tailleFFTW ;

 k++ ; // ATTENTION ! quand on utilise la fftw, on travaille avec la base 2^15 au lieu
 // de 2^30 donc il faut doubler la taille des tableaux.

 tailleFFTW = 1 << k ;

 *inoutA  = fftw_malloc( tailleFFTW * sizeof(fftw_complex) ) ;
 *inoutB  = fftw_malloc( tailleFFTW * sizeof(fftw_complex) ) ;

 *pA  = fftw_plan_dft_1d( tailleFFTW,*inoutA,*inoutA,FFTW_FORWARD,FFTW_ESTIMATE ) ;
 *pB  = fftw_plan_dft_1d( tailleFFTW,*inoutB,*inoutB,FFTW_FORWARD,FFTW_ESTIMATE ) ;
 *p_i = fftw_plan_dft_1d( tailleFFTW,*inoutA,*inoutA,FFTW_BACKWARD,FFTW_ESTIMATE ) ;
}

void destroy_mul_fftw3( fftw_plan* pA , fftw_plan* pB , fftw_plan* p_i ,
		        fftw_complex* inoutA , fftw_complex* inoutB )
{
 fftw_free(inoutA) ;
 fftw_free(inoutB) ;
 fftw_destroy_plan( *pA ) ;
 fftw_destroy_plan( *pB ) ;
 fftw_destroy_plan( *p_i ) ;
}

// Cette routine permet l'initialisation des varaibles ncessaires  l'
// lvation au carr de deux entiers de type Li de taille 2^k en base 2^30.
void init_square_fftw3( long k, fftw_plan* pA , fftw_plan* p_i ,
		     fftw_complex** inoutA )
{
 long tailleFFTW ;

 k++ ; // ATTENTION ! quand on utilise la fftw, on travaille avec la base 2^15 au lieu
 // de 2^30 donc il faut doubler la taille des tableaux.

 tailleFFTW = 1 << k ;

 *inoutA  = fftw_malloc( tailleFFTW * sizeof(fftw_complex) ) ;
 *pA  = fftw_plan_dft_1d( tailleFFTW,*inoutA,*inoutA,FFTW_FORWARD,FFTW_ESTIMATE ) ;
 *p_i = fftw_plan_dft_1d( tailleFFTW,*inoutA,*inoutA,FFTW_BACKWARD,FFTW_ESTIMATE ) ;
}

void destroy_square_fftw3( fftw_plan* pA , fftw_plan* p_i ,
		        fftw_complex* inoutA )
{
 fftw_free(inoutA) ;
 fftw_destroy_plan( *pA ) ;
 fftw_destroy_plan( *p_i ) ;
}

// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
double mul_2P_fftw3_bin_test( Li* prod , Li* a , Li* b , long k ,
		     fftw_plan* pA , fftw_plan* pB , fftw_plan* p_i ,
		     fftw_complex* inoutA , fftw_complex* inoutB )
{
 double tmp0 ;
 double tmp1 ;
 double tmp2 ;
 double tmp3 ;
 double tmp4 ;
 double tmp5 ;
 double tmp6 ;
 double tmp7 ;
 double tmp8 ;
 double tmp9 ;
 double tmp10 ;
 double Ra,Rb,Ia,Ib,RaM,IaM,IbM,RbM ;
 double win,wrn ;
 double i_taille ;
 double error = 0.0 ;
 unsigned long long ll1 , ll2 , lcarry ;
 long i ;
 long t , t2 ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;


 i_taille = 1.0 / taille ;

 for(i=0;i<htaille;i++)
	{
	 t = a->mantisse[i] ;
	 inoutA[i][0] = t & HALF_CACHE32_BIN ;
	 inoutA[i][1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(i=htaille;i<taille;i++)
	{
	 inoutA[i][0] = 0.0 ;
	 inoutA[i][1] = 0.0 ;
	}
 for(i=0;i<htaille;i++)
	{
	 t = b->mantisse[i] ;
	 inoutB[i][0] = t & HALF_CACHE32_BIN ;
	 inoutB[i][1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(i=htaille;i<taille;i++)
	{
	 inoutB[i][0] = 0.0 ;
	 inoutB[i][1] = 0.0 ;
	}

 fftw_execute( *pA ) ;
 fftw_execute( *pB ) ;

 // ------  i = 0
 tmp1 = ( inoutA[0][0] + inoutA[0][1] ) * ( inoutB[0][0] + inoutB[0][1] ) ;
 tmp2 = ( inoutA[0][0] - inoutA[0][1] ) * ( inoutB[0][0] - inoutB[0][1] ) ;
 inoutA[0][0] = 0.5 * ( tmp1 + tmp2 ) ;
 inoutA[0][1] = 0.5 * ( tmp1 - tmp2 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 // taille = 2**(k+1)
 win = -sinus[k+1] ;
 tmp10 = win ;
 tmp9 = sinus[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(i=1;i<htaille;i++)
	{
	 tmp1 = 0.5 * ( inoutA[i][0] + inoutA[taille-i][0] ) ;
	 tmp2 = 0.5 * ( inoutA[i][0] - inoutA[taille-i][0] ) ;
	 tmp3 = 0.5 * ( inoutA[i][1] + inoutA[taille-i][1] ) ;
	 tmp4 = 0.5 * ( inoutA[i][1] - inoutA[taille-i][1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp1 = 0.5 * ( inoutB[i][0] + inoutB[taille-i][0] ) ;
	 tmp2 = 0.5 * ( inoutB[i][0] - inoutB[taille-i][0] ) ;
	 tmp3 = 0.5 * ( inoutB[i][1] + inoutB[taille-i][1] ) ;
	 tmp4 = 0.5 * ( inoutB[i][1] - inoutB[taille-i][1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Rb   =  tmp1 + tmp5 + tmp6 ;
	 Ib   =  tmp4 - tmp7 + tmp8 ;
	 RbM  =  tmp1 - tmp5 - tmp6 ;
	 IbM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = Ra * Rb - Ia * Ib ;
	 Ia   = Ra * Ib + Rb * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = RaM * RbM - IaM * IbM ;
	 IaM  = RaM * IbM + RbM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[i][0] = tmp1 + tmp2 ;
	 inoutA[i][1] = tmp3 + tmp4 ;
	 inoutA[taille-i][0] = tmp1 - tmp2 ;
	 inoutA[taille-i][1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp0 = inoutA[htaille][0] * inoutB[htaille][0] - inoutA[htaille][1] * inoutB[htaille][1] ;
 tmp1 = inoutA[htaille][0] * inoutB[htaille][1] + inoutA[htaille][1] * inoutB[htaille][0] ;
 inoutA[htaille][1] = tmp1 ;
 inoutA[htaille][0] = tmp0 ;

 fftw_execute( *p_i ) ;

 lcarry = 0 ;
 for(i=0;i<taille;i++)
	{
	 tmp1 = inoutA[i][0] * i_taille ;
	 tmp2 = inoutA[i][1] * i_taille ;
	 tmp3 = floor( tmp1 + 0.5 ) ;
	 tmp4 = floor( tmp2 + 0.5 ) ;
	 ll1 = tmp3 ;
	 ll2 = tmp4 ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	 tmp5 = fabs( tmp1 - tmp3 ) ;
	 tmp6 = fabs( tmp2 - tmp4 ) ;
	 if ( tmp5 > error ) error = tmp5 ;
	 if ( tmp6 > error ) error = tmp6 ;
	}

 return error ;

}

// cette routine ne fait pas d'erreur pour k <= 19 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
void mul_2P_fftw3_bin( Li* prod , Li* a , Li* b , long k ,
		     fftw_plan* pA , fftw_plan* pB , fftw_plan* p_i ,
		     fftw_complex* inoutA , fftw_complex* inoutB )
{
 double tmp0 ;
 double tmp1 ;
 double tmp2 ;
 double tmp3 ;
 double tmp4 ;
 double tmp5 ;
 double tmp6 ;
 double tmp7 ;
 double tmp8 ;
 double tmp9 ;
 double tmp10 ;
 double Ra,Rb,Ia,Ib,RaM,IaM,IbM,RbM ;
 double win,wrn ;
 double i_taille ;
 unsigned long long ll1 , ll2 , lcarry ;
 long i ;
 long t , t2 ;
 long htaille = a->taille_mantisse ;
 long taille  = htaille << 1 ;


 i_taille = 1.0 / taille ;

 for(i=0;i<htaille;i++)
	{
	 t = a->mantisse[i] ;
	 inoutA[i][0] = t & HALF_CACHE32_BIN ;
	 inoutA[i][1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(i=htaille;i<taille;i++)
	{
	 inoutA[i][0] = 0.0 ;
	 inoutA[i][1] = 0.0 ;
	}
 for(i=0;i<htaille;i++)
	{
	 t = b->mantisse[i] ;
	 inoutB[i][0] = t & HALF_CACHE32_BIN ;
	 inoutB[i][1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(i=htaille;i<taille;i++)
	{
	 inoutB[i][0] = 0.0 ;
	 inoutB[i][1] = 0.0 ;
	}

 fftw_execute( *pA ) ;
 fftw_execute( *pB ) ;

 // ------  i = 0
 tmp1 = ( inoutA[0][0] + inoutA[0][1] ) * ( inoutB[0][0] + inoutB[0][1] ) ;
 tmp2 = ( inoutA[0][0] - inoutA[0][1] ) * ( inoutB[0][0] - inoutB[0][1] ) ;
 inoutA[0][0] = 0.5 * ( tmp1 + tmp2 ) ;
 inoutA[0][1] = 0.5 * ( tmp1 - tmp2 ) ;
 // ------ initialisation de W(n) = exp( - I * PI * n / taille )
 // taille = 2**(k+1)
 win = -sinus[k+1] ;
 tmp10 = win ;
 tmp9 = sinus[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(i=1;i<htaille;i++)
	{
	 tmp1 = 0.5 * ( inoutA[i][0] + inoutA[taille-i][0] ) ;
	 tmp2 = 0.5 * ( inoutA[i][0] - inoutA[taille-i][0] ) ;
	 tmp3 = 0.5 * ( inoutA[i][1] + inoutA[taille-i][1] ) ;
	 tmp4 = 0.5 * ( inoutA[i][1] - inoutA[taille-i][1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;
	 tmp1 = 0.5 * ( inoutB[i][0] + inoutB[taille-i][0] ) ;
	 tmp2 = 0.5 * ( inoutB[i][0] - inoutB[taille-i][0] ) ;
	 tmp3 = 0.5 * ( inoutB[i][1] + inoutB[taille-i][1] ) ;
	 tmp4 = 0.5 * ( inoutB[i][1] - inoutB[taille-i][1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Rb   =  tmp1 + tmp5 + tmp6 ;
	 Ib   =  tmp4 - tmp7 + tmp8 ;
	 RbM  =  tmp1 - tmp5 - tmp6 ;
	 IbM  = -tmp4 - tmp7 + tmp8 ;
	 tmp0 = Ra * Rb - Ia * Ib ;
	 Ia   = Ra * Ib + Rb * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = RaM * RbM - IaM * IbM ;
	 IaM  = RaM * IbM + RbM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[i][0] = tmp1 + tmp2 ;
	 inoutA[i][1] = tmp3 + tmp4 ;
	 inoutA[taille-i][0] = tmp1 - tmp2 ;
	 inoutA[taille-i][1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp0 = inoutA[htaille][0] * inoutB[htaille][0] - inoutA[htaille][1] * inoutB[htaille][1] ;
 tmp1 = inoutA[htaille][0] * inoutB[htaille][1] + inoutA[htaille][1] * inoutB[htaille][0] ;
 inoutA[htaille][1] = tmp1 ;
 inoutA[htaille][0] = tmp0 ;

 fftw_execute( *p_i ) ;

 lcarry = 0 ;
 for(i=0;i<taille;i++)
	{
	 ll1 = floor( inoutA[i][0] * i_taille + 0.5 ) ;
	 ll2 = floor( inoutA[i][1] * i_taille + 0.5 ) ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	}
}

// cette routine ne fait pas d'erreur pour k <= 20 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
double square_2P_fftw3_test( Li* prod , Li* a , long k ,
		     fftw_plan* pA , fftw_plan* p_i ,
		     fftw_complex* inoutA )
{
 double tmp0 ;
 double tmp1 ;
 double tmp2 ;
 double tmp3 ;
 double tmp4 ;
 double tmp5 ;
 double tmp6 ;
 double tmp7 ;
 double tmp8 ;
 double tmp9 ;
 double tmp10 ;
 double Ra,Ia,RaM,IaM ;
 double win,wrn ;
 double error=0.0 ;
 double i_taille ;
 long i , t , t2 ;
 long taille  = (a->taille_mantisse) << 1 ;
 long htaille = taille >> 1 ;
 unsigned long long ll1 , ll2 , lcarry ;

 i_taille = 1.0 / taille ;

 for(i=0;i<htaille;i++)
	{
	 t = a->mantisse[i] ;
	 inoutA[i][0] = t & HALF_CACHE32_BIN ;
	 inoutA[i][1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(i=htaille;i<taille;i++)
	{
	 inoutA[i][0] = 0.0 ;
	 inoutA[i][1] = 0.0 ;
	}

 fftw_execute( *pA ) ;

 // ------  i = 0
 tmp0 = inoutA[0][0] + inoutA[0][1] ;
 tmp1 = inoutA[0][0] - inoutA[0][1] ;
 tmp2 = tmp0 * tmp0 ;
 tmp3 = tmp1 * tmp1 ;
 inoutA[0][0] = 0.5 * ( tmp2 + tmp3 ) ;
 inoutA[0][1] = 0.5 * ( tmp2 - tmp3 ) ;
 // ------ initialisation de W(n) = exp( I * PI * n / taille )
 // taille = 2**(k+1)
 win = -sinus[k+1] ;
 tmp10 = win ;
 tmp9 = sinus[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(i=1;i<htaille;i++)
	{
	 tmp1 = 0.5 * ( inoutA[i][0] + inoutA[taille-i][0] ) ;
	 tmp2 = 0.5 * ( inoutA[i][0] - inoutA[taille-i][0] ) ;
	 tmp3 = 0.5 * ( inoutA[i][1] + inoutA[taille-i][1] ) ;
	 tmp4 = 0.5 * ( inoutA[i][1] - inoutA[taille-i][1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;

	 tmp0 = ( Ra - Ia ) * ( Ra + Ia ) ;
	 Ia   = 2.0 * Ra * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = ( RaM - IaM ) * ( RaM + IaM ) ;
	 IaM  = 2.0 * RaM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[i][0] = tmp1 + tmp2 ;
	 inoutA[i][1] = tmp3 + tmp4 ;
	 inoutA[taille-i][0] = tmp1 - tmp2 ;
	 inoutA[taille-i][1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp3 = inoutA[htaille][0] ;
 tmp4 = inoutA[htaille][1] ;
 tmp0 = ( tmp3 - tmp4 ) * (tmp3 + tmp4 ) ;
 tmp1 = 2.0 * tmp3 *tmp4 ;
 inoutA[htaille][1] = tmp1 ;
 inoutA[htaille][0] = tmp0 ;


 fftw_execute( *p_i ) ;

 lcarry = 0 ;
 for(i=0;i<taille;i++)
	{
	 tmp1 = inoutA[i][0] * i_taille ;
	 tmp2 = inoutA[i][1] * i_taille ;
	 tmp3 = floor( tmp1 + 0.5 ) ;
	 tmp4 = floor( tmp2 + 0.5 ) ;
	 ll1 = tmp3 ;
	 ll2 = tmp4 ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	 tmp5 = fabs( tmp1 - tmp3 ) ;
	 tmp6 = fabs( tmp2 - tmp4 ) ;
	 if ( tmp5 > error ) error = tmp5 ;
	 if ( tmp6 > error ) error = tmp6 ;
	}
 return error ;
}

// cette routine ne fait pas d'erreur pour k <= 20 o
// taille = 2^k
// a = a(0) + a(1) B + ... + a(taille-1) B^(taille-1)
// B = 2^30
void square_2P_fftw3( Li* prod , Li* a , long k ,
		     fftw_plan* pA , fftw_plan* p_i ,
		     fftw_complex* inoutA )
{
 double tmp0 ;
 double tmp1 ;
 double tmp2 ;
 double tmp3 ;
 double tmp4 ;
 double tmp5 ;
 double tmp6 ;
 double tmp7 ;
 double tmp8 ;
 double tmp9 ;
 double tmp10 ;
 double Ra,Ia,RaM,IaM ;
 double win,wrn ;
 double i_taille ;
 long i , t , t2 ;
 long taille  = (a->taille_mantisse) << 1 ;
 long htaille = taille >> 1 ;
 unsigned long long ll1 , ll2 , lcarry ;


 i_taille = 1.0 / taille ;

 for(i=0;i<htaille;i++)
	{
	 t = a->mantisse[i] ;
	 inoutA[i][0] = t & HALF_CACHE32_BIN ;
	 inoutA[i][1] = t >> HALF_DECALAGE32_BIN ;
	}
 for(i=htaille;i<taille;i++)
	{
	 inoutA[i][0] = 0.0 ;
	 inoutA[i][1] = 0.0 ;
	}

 fftw_execute( *pA ) ;

 // ------  i = 0
 tmp0 = inoutA[0][0] + inoutA[0][1] ;
 tmp1 = inoutA[0][0] - inoutA[0][1] ;
 tmp2 = tmp0 * tmp0 ;
 tmp3 = tmp1 * tmp1 ;
 inoutA[0][0] = 0.5 * ( tmp2 + tmp3 ) ;
 inoutA[0][1] = 0.5 * ( tmp2 - tmp3 ) ;
 // ------ initialisation de W(n) = exp( I * PI * n / taille )
 // taille = 2**(k+1)
 win = -sinus[k+1] ;
 tmp10 = win ;
 tmp9 = sinus[k+2] ;
 tmp9 = tmp9*tmp9 ;
 tmp9 = tmp9 + tmp9 ;
 wrn = 1.0 - tmp9 ;

 for(i=1;i<htaille;i++)
	{
	 tmp1 = 0.5 * ( inoutA[i][0] + inoutA[taille-i][0] ) ;
	 tmp2 = 0.5 * ( inoutA[i][0] - inoutA[taille-i][0] ) ;
	 tmp3 = 0.5 * ( inoutA[i][1] + inoutA[taille-i][1] ) ;
	 tmp4 = 0.5 * ( inoutA[i][1] - inoutA[taille-i][1] ) ;
	 tmp5 = wrn * tmp3 ;
	 tmp6 = win * tmp2 ;
	 tmp7 = wrn * tmp2 ;
	 tmp8 = win * tmp3 ;
	 Ra   =  tmp1 + tmp5 + tmp6 ;
	 Ia   =  tmp4 - tmp7 + tmp8 ;
	 RaM  =  tmp1 - tmp5 - tmp6 ;
	 IaM  = -tmp4 - tmp7 + tmp8 ;

	 tmp0 = ( Ra - Ia ) * ( Ra + Ia ) ;
	 Ia   = 2.0 * Ra * Ia ;
	 Ra   = tmp0 ;
	 tmp0 = ( RaM - IaM ) * ( RaM + IaM ) ;
	 IaM  = 2.0 * RaM * IaM ;
	 RaM  = tmp0 ;
	 tmp6 = 0.5 * ( Ia + IaM ) ;
	 tmp5 = 0.5 * ( Ra - RaM ) ;
	 tmp1 = 0.5 * ( Ra + RaM ) ;
	 tmp2 = win * tmp5 - wrn * tmp6 ;
	 tmp3 = wrn * tmp5 + win * tmp6 ;
	 tmp4 = 0.5 * ( Ia - IaM ) ;
	 inoutA[i][0] = tmp1 + tmp2 ;
	 inoutA[i][1] = tmp3 + tmp4 ;
	 inoutA[taille-i][0] = tmp1 - tmp2 ;
	 inoutA[taille-i][1] = tmp3 - tmp4 ;
	 tmp0 = wrn - ( tmp9 * wrn + tmp10 * win ) ;
	 win  = win - ( tmp9 * win - tmp10 * wrn ) ;
	 wrn  = tmp0 ;
	}

 // ------  i = P = taille / 2
 tmp3 = inoutA[htaille][0] ;
 tmp4 = inoutA[htaille][1] ;
 tmp0 = ( tmp3 - tmp4 ) * (tmp3 + tmp4 ) ;
 tmp1 = 2.0 * tmp3 *tmp4 ;
 inoutA[htaille][1] = tmp1 ;
 inoutA[htaille][0] = tmp0 ;

 fftw_execute( *p_i ) ;

 lcarry = 0 ;
 for(i=0;i<taille;i++)
	{
	 ll1 = floor( inoutA[i][0] * i_taille + 0.5 ) ;
	 ll2 = floor( inoutA[i][1] * i_taille + 0.5 ) ;
	 ll1 = ll1 + lcarry ;
	 t   = ll1 & HALF_CACHE32_BIN ;
	 ll2 = ll2 + (ll1 >> HALF_DECALAGE32_BIN ) ;
	 t2  = ll2 & HALF_CACHE32_BIN ;
	 lcarry = ll2 >> HALF_DECALAGE32_BIN ;
	 prod->mantisse[i] = t + ( t2 << HALF_DECALAGE32_BIN ) ;
	}
}

